#include "MeshLoader.h"
#include <algorithm>
using namespace std;

std::ostream& operator << (std::ostream& out, const std::vector <int>& elem) {
	for (auto& i : elem) {
		out << i << " ";
	}
	out << endl;
	return out;
}

std::ostream& operator << (std::ostream& out, const std::set<int>& elem) {
	for (auto& i : elem) {
		out << i << " ";
	}
	out << endl;
	return out;
}

std::ostream& operator << (std::ostream& out, const std::vector<Node>& elem) {
	for_each(elem.begin(), elem.end(), [&](const auto& i) { //!!! Почему нет const
		out << "   ID node: " << i.ID_Node << std::endl;
		out << "   Coordinates: ";
		for_each(i.Coordin.begin(), i.Coordin.end(), [&](const auto& j) { //!!! Почему нет const
			out << j << " ";
			});
		out << std::endl << "   Flag: " << i.flag << std::endl << std::endl;
		});
	return out;
}

std::ostream& operator << (std::ostream& out, const std::set <FiniteElement>& elem) {
	for_each(elem.begin(), elem.end(), [&](const auto& i) { //!!! Почему нет const
		out << "   ID_KE: " << i.ID_KE << std::endl;
		out << "   ID_Material: " << i.ID_Material << std::endl;
		out << "   ID_Nodes: ";
		for_each(i.ID_Nodes.begin(), i.ID_Nodes.end(), [&](const auto& j) { //!!! Почему нет const
			out << j << " ";
			});
		out << std::endl << std::endl;
		});
	return out;
}
std::ostream& operator << (std::ostream& out, const std::vector <FiniteElement>& elem) {
	for_each(elem.begin(), elem.end(), [&](const auto& i) { //!!! Почему нет const
		out << "   ID_KE: " << i.ID_KE << std::endl;
		out << "   ID_Material: " << i.ID_Material << std::endl;
		out << "   ID_Nodes: ";
		for_each(i.ID_Nodes.begin(), i.ID_Nodes.end(), [&](const auto& j) { //!!! Почему нет const
			out << j << " ";
			});
		out << std::endl << std::endl;
		});
	return out;
}

std::ostream& operator << (std::ostream& out, const std::vector <BoundaryFiniteElement>& elem) {
	for_each(elem.begin(), elem.end(), [&](const auto& i) { //!!! Почему нет const
		out << "   ID_Element: " << i.ID_Element << std::endl;
		out << "   ID_Border: " << i.ID_Border << std::endl;
		out << "   ID_Nodes: ";
		for_each(i.ID_Nodes.begin(), i.ID_Nodes.end(), [&](const auto& j) { //!!! Почему нет const
			out << j << " ";
			});
		out << std::endl << std::endl;
		});
	return out;
}

std::ostream& operator << (std::ostream& out, const std::vector<std::set<int>>& elem) {
	int n = 1;
	for_each(elem.begin(), elem.end(), [&](const auto& i) { //!!! Почему нет const
		out << "Container_" << n << endl;
		out << i;
		++n;
		});
	return out;
}

FiniteElement& MeshLoader::operator[](int k) {
	auto elem = m_FiniteElement.begin();
	while (elem->ID_KE != k) {
		elem++;
	}
	return *elem;
}

std::vector<Node>& MeshLoader::vector_Node() {
	return m_Node; 
}

std::vector<FiniteElement>& MeshLoader::vector_FiniteElement() { 
	return m_FiniteElement; 
}

std::vector<BoundaryFiniteElement>& MeshLoader::vector_BoundaryFiniteElement() { 
	return m_BoundaryFiniteElement; 
}

//!!! Все контейнеры-результаты должны возвращать ID, а не сами КЭ и узлы (НЕ ИСПРАВЛЕНО)
//!!! std::vector<FiniteElement>* -> std::vector<int>

std::vector<int> MeshLoader::FiniteElementByThreeNodes(int id1, int id2, int id3) {
	std::vector<int> result;  //!!! Зачем тут указатель? Где delete тогда?
	auto a = m_FiniteElement.begin();

	while (a != m_FiniteElement.end()) {
		a = find_if(a, m_FiniteElement.end(), [id1, id2, id3](FiniteElement s) {
			return (find(s.ID_Nodes.begin(), s.ID_Nodes.end(), id1) != s.ID_Nodes.end() &&
				find(s.ID_Nodes.begin(), s.ID_Nodes.end(), id2) != s.ID_Nodes.end() &&
				find(s.ID_Nodes.begin(), s.ID_Nodes.end(), id3) != s.ID_Nodes.end());
			});
		if (a != m_FiniteElement.end()) {
			result.push_back(a->ID_KE);
			a++;
		}
	}

	return result;

}

//!!! std::vector<FiniteElement>* -> std::vector<int>

std::vector<int> MeshLoader::FiniteElementByTwoNodes(int id1, int id2) {
	std::vector<int> result; //!!! Зачем тут указатель? Где delete тогда?
	auto a = m_FiniteElement.begin();

	while (a != m_FiniteElement.end()) {
		a = find_if(a, m_FiniteElement.end(), [id1, id2](FiniteElement& elem) { //!!! Копирование s. Имена должны быть осмысленные
			return (find(elem.ID_Nodes.begin(), elem.ID_Nodes.end(), id1) != elem.ID_Nodes.end() &&
				find(elem.ID_Nodes.begin(), elem.ID_Nodes.end(), id2) != elem.ID_Nodes.end());
			});
		if (a != m_FiniteElement.end()) {
			result.push_back(a->ID_KE);
			a++;
		}
	}

	return result;

}

//!!! std::set<Node>* -> std::set<int>

std::set<int> MeshLoader::BoundaryNodes(int id) {
	std::set<int> result;  //!!! Зачем тут указатель? Где delete тогда?

	for_each(m_BoundaryFiniteElement.begin(), m_BoundaryFiniteElement.end(), [&](const auto& elem1) { //!!! Почему нет const 
		if (elem1.ID_Border == id) {
			for_each(elem1.ID_Nodes.begin(), elem1.ID_Nodes.end(), [&](const auto& elem2) {//!!! Почему нет const 

					result.insert((m_Node.at(elem2 - 1)).ID_Node);
				});
		}
		});

	return result;
}

//!!! ...
std::set<int> MeshLoader::ContainerKE_by_id_border(int id_border) {
	std::set<int> result; //!!! ????????????

	for_each(m_BoundaryFiniteElement.begin(), m_BoundaryFiniteElement.end(), [&](auto& elem) {
		if (elem.ID_Border == id_border) {
		    
			//!!! Для каждого граничено КЭ обходятся все тетраэдры? Что тут делается?
			//!!! Как обходились, так и обходятся - ничего не изменилось
			auto a= FiniteElementByThreeNodes(elem.ID_Nodes.at(0), elem.ID_Nodes.at(1), elem.ID_Nodes.at(2));
			for (auto& i : a) {
				result.insert(i);
			}
		}
		});
	return result;
}

//!!! std::vector<BoundaryFiniteElement>* -> std::vector<int>
std::vector<int> MeshLoader::BoundaryKE(int id) {
	std::vector<int> result; //!!! ??????????????????????????

	for_each(m_BoundaryFiniteElement.begin(), m_BoundaryFiniteElement.end(), [&](auto& elem) { //!!! Копирование elem 
		if (elem.ID_Border == id) {
			result.push_back(elem.ID_Element);
		}
		});

	return result;
}

void MeshLoader::NewNodes() {

	unordered_set<Edge, EdgeHash> List_edge;
	for (auto& i : m_FiniteElement) {
		vector<int> a = i.ID_Nodes;
		for (auto elem1 = a.begin(); elem1 < a.end(); elem1++) {
			for (auto elem2 = elem1 + 1; elem2 < a.end(); elem2++) {
					Edge ed(*elem1, *elem2);
					auto el = List_edge.find(ed);
					if (el == List_edge.end()) {
						Node el;
						el.ID_Node = m_Node.size() + 1;
						el.flag = false;
						for (int j = 0; j < m_Node.begin()->Coordin.size(); j++) {
							el.Coordin.push_back((m_Node.at(*elem1 - 1).Coordin.at(j) + m_Node.at(*elem2 - 1).Coordin.at(j)) / 2);
						}
						m_Node.push_back(el);
						i.ID_Nodes.push_back(el.ID_Node);
						ed.middle_id = el.ID_Node;
						List_edge.insert(ed);
					}
					else {
						 //!!! Дублирование кода - поиск 2 раза 
						i.ID_Nodes.push_back(el->middle_id);
					}
			}
		}
	}
	for (auto& i : m_BoundaryFiniteElement) {
		vector<int> a = i.ID_Nodes;
		for (auto elem1 = a.begin(); elem1 < a.end(); elem1++) {
			for (auto elem2 = elem1 + 1; elem2 < a.end(); elem2++) {
				Edge ed(*elem1, *elem2);
				if (List_edge.find(ed) != List_edge.end()) {
					auto el = List_edge.find(ed);
					i.ID_Nodes.push_back(el->middle_id);
				}
			}
		}
	}

}

//!!! std::vector<std::set<Node>>* -> std::vector<std::set<int>>
std::vector<std::set<int>> MeshLoader::Container_of_containers_of_neighboring_nodes() {
	std::vector<std::set<int>> result(m_Node.size()); //!!! ?????????????????????????????????
	for (auto& elem : m_FiniteElement) {
		for (auto elem1 = elem.ID_Nodes.begin(); elem1 < elem.ID_Nodes.end(); elem1++) {
			for (auto elem2 = elem1 + 1; elem2 < elem.ID_Nodes.end(); elem2++) {
				result[*elem1-1].insert(*elem2);
				result[*elem2-1].insert(*elem1);
			}
		}
	}
	return result;
}


void MeshLoader::print_Node() {
	cout << m_Node;
}

void MeshLoader::print_FiniteElement() {
	cout << m_FiniteElement;
}

void MeshLoader::print_BoundaryFiniteElement() {
	cout << m_BoundaryFiniteElement;
}

